from netCDF4 import Dataset  # require netcdf4 being installed, not NetCDF4
from numpy import ma

from viirs_tools.assimilator import reading_helpers as rh


def read_npp_viaes_l1(path: str) -> dict[str, ma.MaskedArray]:
    """Read VIIRS I-band imagery product (VIAES_L1)

    Args:
        path : path to the desired file

    Returns:
        Datasets with reflectance, radiance,
            brightness temperature data
            as masked np-arrays
    """
    data = {}
    with Dataset(path, "r") as ifile:
        # get reflectance data
        for i in range(1, 4):
            data[f"refi{i}"] = rh.read_ref(f"I{i}", ifile)

        # get brightness temperature data
        for i in range(4, 6):
            data[f"bti{i}"] = rh.read_bt(f"I{i}", ifile)

        # get radiance data
        for i in range(1, 6):
            data[f"radi{i}"] = rh.read_rad(f"I{i}", ifile)

    return data


def read_npp_vmaes_l1(
    path: str,
) -> tuple[dict[str, ma.MaskedArray], dict[str, ma.MaskedArray]]:
    """Read VIIRS M-band imagery product (VMAES_L1)

    Args:
        path : path to the desired file

    Returns:
        Datasets with reflectance, radiance,
            brightness temperature data, geo-reference
            as masked np-arrays
    """
    data = {}
    geo = {}
    with Dataset(path, "r") as mfile:
        # get reflectance data
        for i in range(1, 12):
            data[f"refm{i}"] = rh.read_ref(f"M{i}", mfile)

        # get brightness temperature data
        for i in range(12, 17):
            data[f"btm{i}"] = rh.read_bt(f"M{i}", mfile)

        # get radiance data
        for i in range(1, 17):
            data[f"radm{i}"] = rh.read_rad(f"M{i}", mfile)

        # get geo data
        geo["lat"] = rh.read_lat(mfile)
        geo["lon"] = rh.read_lon(mfile)
        geo["solza"] = rh.read_solza(mfile)
        geo["solaa"] = rh.read_solaa(mfile)
        geo["satza"] = rh.read_satza(mfile)
        geo["sataa"] = rh.read_sataa(mfile)

    return data, geo


def read_npp_cldmsk_l2(path: str) -> dict[str, ma.MaskedArray]:
    """Read VIIRS Cloud Mask product (CLDMSK_L2)

    Args:
        path : path to the desired file

    Returns:
        Datasets clear_conf and integer cloud mask as masked np-arrays
    """
    data = {}
    with Dataset(path, "r") as file:
        data["clear_conf"] = file.groups["geophysical_data"].variables["Clear_Sky_Confidence"][::]
        data["cloud_mask"] = file.groups["geophysical_data"].variables["Integer_Cloud_Mask"][::]
    return data



from collections.abc import Callable
from enum import Enum
from types import MappingProxyType

from viirs_tools.algs import cloud, index, lst, night, utils, water
from viirs_tools.utils.types import AlgEnum


class AlgsCloud(AlgEnum):
    VIBCM_DAY = "AlgsCloud.VIBCM_DAY: VIIRS I-Band Cloud Mask, day time, [default]"
    VIFCM_DAY = "AlgsCloud.VIFCM: VIIRS Active Fire Cloud Mask, day time"
    VIFCM_NIGHT = "AlgsCloud.VIFCM: VIIRS Active Fire Cloud Mask, night time"


class AlgsIndex(AlgEnum):
    NDVI = "AlgsIndex.NDVI: NDVI, day time, [default]"
    NDSI = "AlgsIndex.NDSI: NDSI, day time"


class AlgsLST(AlgEnum):
    MONO_WINDOW_I05 = "AlgsLST.MONO_WINDOW_I05: Mono-window at I05 band, day time, [default]"
    MONO_WINDOW_M15 = "AlgsLST.MONO_WINDOW_M15: Mono-window at M15 band, day time"
    MONO_WINDOW_M16 = "AlgsLST.MONO_WINDOW_M16: Mono-window at M16 band, day time"


class AlgsNight(AlgEnum):
    NAIVE = "AlgsNight.NAIVE: Naive, any time, [default]"


class AlgsWater(AlgEnum):
    WBODIES_DAY = "AlgsWater.WBODIES_DAY: Water bodies, day time, [default]"


class AlgsUtils(AlgEnum):
    MERGE_DAY_NIGHT = "AlgsUtils.MERGE_DAY_NIGHT: Merge data by day-night mask, [default]"


class Runner:
    _IMPL_AlgsIndex = MappingProxyType({AlgsIndex.NDVI: index.ndvi, AlgsIndex.NDSI: index.ndsi})
    _IMPL_AlgsNight = MappingProxyType({AlgsNight.NAIVE: night.naive})
    _IMPL_AlgsCloud = MappingProxyType(
        {
            AlgsCloud.VIBCM_DAY: cloud.vibcm_day,
            AlgsCloud.VIFCM_DAY: cloud.vifcm_day,
            AlgsCloud.VIFCM_NIGHT: cloud.vifcm_night,
        }
    )
    _IMPL_AlgsLST = MappingProxyType(
        {
            AlgsLST.MONO_WINDOW_I05: lst.mono_window_i05,
            AlgsLST.MONO_WINDOW_M15: lst.mono_window_m15,
            AlgsLST.MONO_WINDOW_M16: lst.mono_window_m16,
        }
    )
    _IMPL_AlgsWater = MappingProxyType({AlgsWater.WBODIES_DAY: water.water_bodies_day})
    _IMPL_AlgsUtils = MappingProxyType({AlgsUtils.MERGE_DAY_NIGHT: utils.merge_day_night})

    _IMPLS = MappingProxyType(
        {
            AlgsIndex: _IMPL_AlgsIndex,
            AlgsNight: _IMPL_AlgsNight,
            AlgsCloud: _IMPL_AlgsCloud,
            AlgsLST: _IMPL_AlgsLST,
            AlgsWater: _IMPL_AlgsWater,
            AlgsUtils: _IMPL_AlgsUtils,
        }
    )

    def __init__(self):
        pass

    def _show_algs(self, algs: dict[Enum, Callable]):
        print("<Key>: <Description>")
        for item in algs.items():
            print(item)

    def show_algs_index(self):
        self._show_algs(Runner._IMPL_AlgsIndex)

    def show_algs_night(self):
        self._show_algs(Runner._IMPL_AlgsNight)

    def show_algs_cloud(self):
        self._show_algs(Runner._IMPL_AlgsCloud)

    def show_algs_lst(self):
        self._show_algs(Runner._IMPL_AlgsLST)

    def show_algs_water(self):
        self._show_algs(Runner._IMPL_AlgsWater)

    def show_algs_utils(self):
        self._show_algs(Runner._IMPL_AlgsUtils)

    def show_algs_all(self):
        for impl in Runner._IMPLS.values():
            self._show_algs(impl)
            print()

    def _get_alg(self, impls, algs, alg=None) -> Callable:
        if alg is None:
            alg = next(iter(algs))
        return impls[alg]

    def get_alg_index(self, alg: AlgsIndex | None = None) -> Callable:
        return self._get_alg(Runner._IMPL_AlgsIndex, AlgsIndex, alg=alg)

    def get_alg_night(self, alg: AlgsNight | None = None) -> Callable:
        return self._get_alg(Runner._IMPL_AlgsNight, AlgsNight, alg=alg)

    def get_alg_cloud(self, alg: AlgsCloud | None = None) -> Callable:
        return self._get_alg(Runner._IMPL_AlgsCloud, AlgsCloud, alg=alg)

    def get_alg_lst(self, alg: AlgsLST | None = None) -> Callable:
        return self._get_alg(Runner._IMPL_AlgsLST, AlgsLST, alg=alg)

    def get_alg_water(self, alg: AlgsWater | None = None) -> Callable:
        return self._get_alg(Runner._IMPL_AlgsWater, AlgsWater, alg=alg)

    def get_alg_utils(self, alg: AlgsUtils | None = None) -> Callable:
        return self._get_alg(Runner._IMPL_AlgsUtils, AlgsUtils, alg=alg)



import numpy as np
import xarray as xr

from viirs_tools.utils.types import ArrayLike, _check_data


def _mono_window(
    bt: ArrayLike,
    band_lambda: float,
    ndvi: ArrayLike,
    cmask: ArrayLike | None = None,
) -> ArrayLike:
    """LST retrieval algorithm for day conditions
    Based on the U.Adam, G.Jovanoska (2016).
    Algorithm for Automated Mapping of
        Land Surface Temperature Using LANDSAT 8 Satellite Data

    Args:
        bt : band at ~10-12 um in BT calibration
        band_lambda : wavelength of emitted radiance in used band
            (the peak response or average of the limiting wavelength)
            Unit is um
        ndvi : NDVI in corresponding resolution
        cmask : integer cloud mask, 1 is clear sky pixel

    Returns:
        Array containing LST, can contain NaN values
    """
    if cmask is not None:
        _check_data(bt, ndvi, cmask)
    else:
        _check_data(bt, ndvi)

    bt_c = bt - 273.15  # to Celsius

    ndvi_s = 0.2
    ndvi_v = 0.5
    p_v = ((ndvi - ndvi_s) / (ndvi_v - ndvi_s)) ** 2

    c = 0.005  # represents surface roughness

    e_s = 0.966  # soil emissivity
    e_v = 0.973  # vegetation emissivity
    e_w = 0.991  # water emmisivity

    e_l = e_v * p_v + e_s * (1 - p_v) + c
    e_l = xr.where(ndvi < ndvi_s, e_s, e_l)
    e_l = xr.where(ndvi < 0, e_w, e_l)  # ndvi < 0 indicates water
    e_l = xr.where(ndvi > ndvi_v, e_v, e_l)

    p = 1.438e-2

    lst = bt_c / (1 + (band_lambda * bt_c / p) * np.log(e_l) * 1e-12)
    if cmask is not None:
        lst[cmask == 0] = np.nan
    return lst


def mono_window_i05(bi05: ArrayLike, ndvi: ArrayLike, cmask: ArrayLike | None = None) -> ArrayLike:
    """
        LST retrieval algorithm for day conditions
        Based on the U.Adam, G.Jovanoska (2016).
        Algorithm for Automated Mapping of
            Land Surface Temperature Using LANDSAT 8 Satellite Data
    Args:
        bi05 (ArrayLike) : I05 in BT calibration
        ndvi (ArrayLike) : NDVI in corresponding resolution
        cmask (ArrayLike | None) : integer cloud mask,
            1 is clear sky pixel
    Returns:
        (ArrayLike) : array containing LST,
            Can contain NaN values
    """
    return _mono_window(bi05, (10.5 + 12.4) / 2, ndvi, cmask=cmask)


def mono_window_m15(bm15: ArrayLike, ndvi: ArrayLike, cmask: ArrayLike | None = None) -> ArrayLike:
    """
        LST retrieval algorithm for day conditions
        Based on the U.Adam, G.Jovanoska (2016).
        Algorithm for Automated Mapping of
            Land Surface Temperature Using LANDSAT 8 Satellite Data
    Args:
        bm15 (ArrayLike) : M15 in BT calibration
        ndvi (ArrayLike) : NDVI in corresponding resolution
        cmask: (np.ndarray|xr.Dataset, optional): integer cloud mask,
            1 is clear sky pixel
    Returns:
        (ArrayLike) : array containing LST,
            Can contain NaN values
    """
    return _mono_window(bm15, (10.263 + 11.263) / 2, ndvi, cmask=cmask)


def mono_window_m16(bm16: ArrayLike, ndvi: ArrayLike, cmask: ArrayLike | None = None) -> ArrayLike:
    """
        LST retrieval algorithm for day conditions
        Based on the U.Adam, G.Jovanoska (2016).
        Algorithm for Automated Mapping of
            Land Surface Temperature Using LANDSAT 8 Satellite Data
    Args:
        bm16 (ArrayLike) : M16 in BT calibration
        ndvi (ArrayLike) : NDVI in corresponding resolution
        cmask (ArrayLike | None) : integer cloud mask,
            1 is clear sky pixel
    Returns:
        (np.ndarray, xr.Dataset): array containing LST,
            Can contain NaN values
    """
    return _mono_window(bm16, (11.538 + 12.488) / 2, ndvi, cmask=cmask)



import gc
import os
import re
import time
from datetime import UTC, datetime

import satpy
from pyresample.geometry import AreaDefinition

from viirs_tools.assimilator.assimilator import assimilate
from viirs_tools.assimilator.reading import read_npp_cldmsk_l2

area_by = AreaDefinition(
    area_id="Belarus",
    description="The region of Belarus (WGS 84 / UTM zone 35N Projection)",
    projection="EPSG:32635",
    proj_id="EPSG:32635",
    # width = 700,  # resolution=1000
    # height = 600, # resolution=1000
    width=1400,  # resolution=500
    height=1200,  # resolution=500
    area_extent=(
        200000,
        5650000,
        900000,
        6250000,
    ),
)

DEBUG = True


def log(msg, debug=DEBUG):
    if debug:
        print(msg)


def generate_composits(prefix, i_file, c_file, ig_file, mg_file):
    gc.collect()

    if os.path.exists(prefix):
        os.system(f"rm -rf {prefix}")
    os.mkdir(prefix)

    # load i_file, ig_file, generate i-bands composits
    log("[LOG] Generating I-band data..")
    si = satpy.Scene(filenames=[i_file, ig_file], reader="viirs_l1b")

    is_day = "I01" in si.available_dataset_names()
    if is_day:
        si.load(["I01", "I02", "I03"])
    si.load(["I04", "I05", "i_lat", "i_lon"])

    rsi = si.resample(destination=area_by, resampler="nearest")

    log("[LOG] probing for enough amount of data..")
    mask = rsi["I05"].notnull()
    s = mask.sum()
    if s < 0.1 * mask.shape[0] * mask.shape[1]:
        log("[WARN] Not enough data, skipping..")
        os.rmdir(prefix)
        del rsi
        del si
        gc.collect()
    else:
        if is_day:
            rsi.save_dataset(
                "I01",
                writer="geotiff",
                enhance=False,
                filename=os.path.join(prefix, "i01_ref.tiff"),
            )
            rsi.save_dataset(
                "I02",
                writer="geotiff",
                enhance=False,
                filename=os.path.join(prefix, "i02_ref.tiff"),
            )
            rsi.save_dataset(
                "I03",
                writer="geotiff",
                enhance=False,
                filename=os.path.join(prefix, "i03_ref.tiff"),
            )
        rsi.save_dataset(
            "I04",
            writer="geotiff",
            enhance=False,
            filename=os.path.join(prefix, "i04_bt.tiff"),
        )
        rsi.save_dataset(
            "I05",
            writer="geotiff",
            enhance=False,
            filename=os.path.join(prefix, "i05_bt.tiff"),
        )
        log("[LOG] writing is done")

        del rsi
        del si
        gc.collect()

        # load c_file, mg_file, generate cm composit
        log("[LOG] Generating cloud mask data..")
        s = satpy.Scene(filenames=[mg_file], reader="viirs_l1b")
        s.load(["solar_zenith_angle", "m_lat", "m_lon"])
        d = s["solar_zenith_angle"]
        d.attrs["name"] = "integer_cloud_mask"
        d.attrs["standard_name"] = "integer_cloud_mask"
        d.attrs["long_name"] = "integer cloud mask"
        d.attrs["valid_min"] = 0
        d.attrs["valid_max"] = 1
        d.attrs["scale_factor"] = 1
        d.attrs["units"] = "none"
        d.attrs["file_key"] = ""
        d.attrs["file_units"] = "none"
        d.attrs["reader"] = ""
        cm_h = read_npp_cldmsk_l2(c_file)
        cm = cm_h["cloud_mask"]
        d.data = cm.astype(d.dtype)
        rs = s.resample(destination=area_by, resampler="nearest")
        rs.save_dataset(
            "solar_zenith_angle",
            writer="geotiff",
            enhance=False,
            filename=os.path.join(prefix, "mvcm.tiff"),
        )
        del rs
        del s
        gc.collect()
        log(f"[LOG] {prefix} is done")
        return True
    log(f"[LOG] {prefix} is cancelled")
    return False


def my_assim(path):
    i_files = []
    c_files = []
    mg_files = []
    ig_files = []

    files = os.listdir(path)
    for file in files:
        filepath = os.path.join(path, file)
        if re.match(r"VNP02IMG", file) is not None:
            i_files.append(filepath)
        if re.match(r"CLDMSK", file) is not None:
            c_files.append(filepath)
        if re.match(r"VNP03IMG", file) is not None:
            ig_files.append(filepath)
        if re.match(r"VNP03MOD", file) is not None:
            mg_files.append(filepath)

    k = 1
    for c_file in c_files:
        ts = re.search(r"A\d{7}\.\d{4}", c_file)[0]
        i_file = ""
        mg_file = ""
        ig_file = ""

        for file in i_files:
            if ts in file:
                i_file = file
                break

        for file in mg_files:
            if ts in file:
                mg_file = file
                break

        for file in ig_files:
            if ts in file:
                ig_file = file
                break

        if ig_file == "" or mg_file == "" or i_file == "":
            print(f"[WARN] [{k}/{len(c_files)}] Skipping timestamp {ts}")
        else:
            print(f"[LOG] [{k}/{len(c_files)}] Generating data for {ts}..")
            generate_composits(os.path.join(path, ts), i_file, c_file, ig_file, mg_file)
        k += 1

    print("[LOG] cleanup..")

    for i in i_files:
        os.remove(i)
    for i in c_files:
        os.remove(i)
    for i in mg_files:
        os.remove(i)
    for i in ig_files:
        os.remove(i)

    print(f"[LOG] {path} is done")


names = ["VNP02IMG", "VNP03IMG", "VNP03MOD", "CLDMSK_L2_VIIRS_SNPP"]
BY_BOX = "23.1774,51.256,32.7628,56.172"
sd = datetime(2012, 3, 1, tzinfo=UTC)
ed = datetime(2012, 11, 7, tzinfo=UTC)


t1 = time.time()
assimilate(names, sd, ed, "./2012", BY_BOX, 1, assim_callback=my_assim)
t2 = time.time()

print(f"Done in {t2 - t1}")



import numpy as np
import xarray as xr

from viirs_tools.algs import index
from viirs_tools.utils.types import ArrayLike, _check_data


def vibcm_day(
    ri1: ArrayLike,
    ri2: ArrayLike,
    ri3: ArrayLike,
    bi5: ArrayLike,
    ndsi: ArrayLike | None = None,
) -> ArrayLike:
    """Day reflectance/thermal I-bands cloud test
    Based on the M.Piper, T.Bahr (2015).
    A RAPID CLOUD MASK ALGORITHM FOR SUOMI NPP VIIRS IMAGERY EDRS

    Args:
        ri1 : I01 in reflectance calibration
        ri2 : I02 in reflectance calibration
        ri3 : I03 in reflectance calibration
        bi5 : I05 in BT calibration

    Returns:
        Integer cloud mask, 0 is cloud, 1 is clear pixel
            Can contain NaN values
    """
    if ndsi is None:
        _check_data(ri1, ri2, ri3, bi5)
    else:
        _check_data(ri1, ri2, ri3, bi5, ndsi)

    # Test 1
    cm = xr.where(ri1 > 8, True, False)
    print(cm)

    # Test 2
    mask = index.ndsi(ri1, ri3) > 0.7 if ndsi is None else ndsi > 0.7
    mask = xr.where(ri2 > 11, mask, False)
    cm = xr.where(mask, cm, False)
    print(cm)

    # Test 3
    cm = xr.where(bi5 < 300, cm, False)
    print(cm)

    # Test 4
    ri3_max = np.nanmax(ri3, axis=(-2, -1), keepdims=True)
    cm = xr.where((ri3_max - ri3) * bi5 / 100 < 410, cm, False)
    print(cm)

    # Test 5
    cm = xr.where(ri2 / ri1 < 2, cm, False)
    print(cm)

    # Test 6
    cm = xr.where(ri2 / ri3 > 1, cm, False)
    print(cm)

    return 1 - xr.where(xr.ufuncs.isnan(ri1), np.nan, cm)


def vifcm_day(ri1: ArrayLike, ri2: ArrayLike, bi5: ArrayLike) -> ArrayLike:
    """Day reflectance & termal I-bands cloud test
    Based on the W.Schroeder, P.Oliva, L.Giglio, I.A.Csiszar (2014).
    The New VIIRS 375 m active fire detection data product:
        Algorithm description and initial assessment

    Args:
        ri1 : I01 in reflectance calibration
        ri2 : I02 in reflectance calibration
        bi5 : I05 in BT calibration

    Returns:
        Integer cloud mask, 0 is cloud, 1 is clear pixel
            Can contain NaN values
    """
    _check_data(ri1, ri2, bi5)

    # Test 1
    cm = xr.where(bi5 < 265, True, False)

    sum_ri = ri1 + ri2

    # Test 2
    cm = xr.where((sum_ri > 90) & (bi5 < 295), True, cm)

    # Test 3
    cm = xr.where((sum_ri > 70) & (bi5 < 285), True, cm)

    return 1 - xr.where(xr.ufuncs.isnan(ri1), np.nan, cm)


def vifcm_night(bi4: ArrayLike, bi5: ArrayLike, nmask: ArrayLike | None = None) -> ArrayLike:
    """Night termal I-bands cloud test
    Based on the W.Schroeder, P.Oliva, L.Giglio, I.A.Csiszar (2014).
    The New VIIRS 375 m active fire detection data product:
        Algorithm description and initial assessment

    Args:
        bi4 : I04 in BT calibration
        bi5 : I05 in BT calibration
        nmask : Day/night mask (1 is night)

    Returns:
        Integer cloud mask, 0 is cloud, 1 is clear pixel
            Can contain NaN values
    """
    if nmask is not None:
        _check_data(bi4, bi5, nmask)
    else:
        _check_data(bi4, bi5)

    cm = (bi5 < 265) & (bi4 < 295)

    if nmask is not None:
        cm = xr.where(nmask, cm, np.nan)

    return 1 - xr.where(xr.ufuncs.isnan(bi4), np.nan, cm)



import numpy as np
import xarray as xr

IMAGE_SHAPE = (3, 3)


def get_data_np(shape):
    def _get_band(shape):
        return np.random.rand(*shape) * 100

    mask = _get_band(shape) < 50
    ri1 = _get_band(shape)
    ri2 = _get_band(shape)
    ri3 = _get_band(shape)
    bi4 = _get_band(shape)
    bi5 = _get_band(shape)

    ri1[mask] = np.nan
    ri2[mask] = np.nan
    ri3[mask] = np.nan
    bi4[mask] = np.nan
    bi5[mask] = np.nan

    return ri1, ri2, ri3, bi4, bi5


def _np2xr(arr):
    dims = ("time", "x", "y") if len(arr.shape) == 3 else ("x", "y")
    coords = {"x": np.arange(arr.shape[-2]), "y": np.arange(arr.shape[-1])}
    if len(arr.shape) == 3:
        coords["time"] = np.arange(arr.shape[0])
    return xr.DataArray(arr, dims=dims, coords=coords)


def get_data_xr(shape):
    ri1, ri2, ri3, bi4, bi5 = get_data_np(shape)

    return (
        _np2xr(ri1),
        _np2xr(ri2),
        _np2xr(ri3),
        _np2xr(bi4),
        _np2xr(bi5),
    )


def get_np_from_list(data: list) -> np.ndarray:
    image = np.array(data)
    dims = len(image.shape)
    assert 1 <= dims < 3
    if len(image.shape) == 1:
        image = image[..., None]
    return image


def get_np_seq_from_list(data: list, layers: int = 3) -> np.ndarray:
    layer = get_np_from_list(data)
    return np.stack(layers * [layer])


def get_xr_from_list(data: list) -> xr.DataArray:
    return _np2xr(get_np_from_list(data))


def get_xr_seq_from_list(data: list, layers: int = 3) -> xr.DataArray:
    return _np2xr(get_np_seq_from_list(data, layers=layers))



from viirs_tools.utils.types import ArrayLike, _check_data


def ndvi(nir: ArrayLike, r: ArrayLike) -> ArrayLike:
    """Get Normalized Difference Vegetation Index from reflectance bands
    Most common use is (I2, I1) or (M7, M5) as (NIR, R)

    Args:
        nir : near-infrared reflectance band
        r : red reflectance band

    Returns:
        NDVI index for each pixel, can contain NaN values
    """
    _check_data(nir, r)

    return (nir - r) / (nir + r)


def ndsi(ri1: ArrayLike, ri3: ArrayLike) -> ArrayLike:
    """Get Normalized Difference Snow Index from I01 and I03 reflectance bands

    Args:
        ri1 : I01 in reflectance calibration
        ri3 : I03 in reflectance calibration

    Returns:
        NDVI index for each pixel, can contain NaN values
    """
    _check_data(ri1, ri3)

    return (ri1 - ri3) / (ri1 + ri3)



import numpy as np
import pytest

from tests.algs.utils import (
    IMAGE_SHAPE,
    get_data_np,
    get_data_xr,
    get_np_from_list,
    get_np_seq_from_list,
    get_xr_from_list,
    get_xr_seq_from_list,
)
from viirs_tools.algs import cloud


class TestVibcmDay:
    def test_smoke_np(self):
        def _test(shape: tuple[int, ...]):
            ri1, ri2, ri3, bi4, bi5 = get_data_np(shape)
            cloud.vibcm_day(ri1, ri2, ri3, bi4, bi5)

        _test(IMAGE_SHAPE)
        _test((2, *IMAGE_SHAPE))

    def test_smoke_xr(self):
        def _test(shape: tuple[int, ...]):
            ri1, ri2, ri3, bi4, bi5 = get_data_xr(shape)
            cloud.vibcm_day(ri1, ri2, ri3, bi4, bi5)

        _test(IMAGE_SHAPE)
        _test((2, *IMAGE_SHAPE))

    @pytest.mark.parametrize(
        ("ri1", "ri2", "ri3", "bi5", "expected"),
        [
            (
                # ri1
                [np.nan, 100],
                # ri2
                [np.nan, 20],
                # ri3
                [np.nan, 10],
                # bi5
                [np.nan, 200],
                # expected
                [np.nan, 0],
            ),
            (
                # ri1
                [100, 120, 7, 9, 100],
                # ri2
                [20, 30, 12, 30, 20],
                # ri3
                [10, 25, 1, 1, 40],
                # bi5
                [400, 200, 200, 200, 200],
                # expected
                [1, 1, 1, 1, 1],
            ),
        ],
    )
    def test_alg(self, ri1, ri2, ri3, bi5, expected):
        mask = cloud.vibcm_day(get_np_from_list(ri1), get_np_from_list(ri2), get_np_from_list(ri3), get_np_from_list(bi5))
        assert np.allclose(mask, get_np_from_list(expected), equal_nan=True)

        mask = cloud.vibcm_day(get_np_seq_from_list(ri1), get_np_seq_from_list(ri2), get_np_seq_from_list(ri3), get_np_seq_from_list(bi5))
        assert np.allclose(mask, get_np_seq_from_list(expected), equal_nan=True)

        mask = cloud.vibcm_day(get_xr_from_list(ri1), get_xr_from_list(ri2), get_xr_from_list(ri3), get_xr_from_list(bi5))
        assert mask.equals(get_xr_from_list(expected))

        mask = cloud.vibcm_day(get_xr_seq_from_list(ri1), get_xr_seq_from_list(ri2), get_xr_seq_from_list(ri3), get_xr_seq_from_list(bi5))
        assert mask.equals(get_xr_seq_from_list(expected))


class TestVifcmDay:
    def test_smoke_np(self):
        def _test(shape: tuple[int, ...]):
            ri1, ri2, _, _, bi5 = get_data_np(shape)
            cloud.vifcm_day(ri1, ri2, bi5)

        _test(IMAGE_SHAPE)
        _test((2, *IMAGE_SHAPE))

    def test_smoke_xr(self):
        def _test(shape: tuple[int, ...]):
            ri1, ri2, _, _, bi5 = get_data_xr(shape)
            cloud.vifcm_day(ri1, ri2, bi5)

        _test(IMAGE_SHAPE)
        _test((2, *IMAGE_SHAPE))

    @pytest.mark.parametrize(
        ("ri1", "ri2", "bi5", "expected"),
        [
            (  # missing values
                # ri1
                [np.nan],
                # ri2
                [np.nan],
                # bi5
                [np.nan],
                # expected
                [np.nan],
            ),
            (  # all tests failed/passed
                # ri1
                [10, 50],
                # ri2
                [10, 50],
                # bi5
                [300, 260],
                # expected
                [1, 0],
            ),
            (  # single test passed
                # ri1
                [10, 50, 40],
                # ri2
                [10, 50, 40],
                # bi5
                [150, 290, 280],
                # expected
                [0, 0, 0],
            ),
        ],
    )
    def test_alg(self, ri1, ri2, bi5, expected):
        mask = cloud.vifcm_day(get_np_from_list(ri1), get_np_from_list(ri2), get_np_from_list(bi5))
        assert np.allclose(mask, get_np_from_list(expected), equal_nan=True)

        mask = cloud.vifcm_day(get_np_seq_from_list(ri1), get_np_seq_from_list(ri2), get_np_seq_from_list(bi5))
        assert np.allclose(mask, get_np_seq_from_list(expected), equal_nan=True)

        mask = cloud.vifcm_day(get_xr_from_list(ri1), get_xr_from_list(ri2), get_xr_from_list(bi5))
        assert mask.equals(get_xr_from_list(expected))

        mask = cloud.vifcm_day(get_xr_seq_from_list(ri1), get_xr_seq_from_list(ri2), get_xr_seq_from_list(bi5))
        assert mask.equals(get_xr_seq_from_list(expected))


class TestVifcmNight:
    def test_smoke_np(self):
        def _test(shape: tuple[int, ...]):
            _, _, _, bi4, bi5 = get_data_np(shape)
            cloud.vifcm_night(bi4, bi5)

        _test(IMAGE_SHAPE)
        _test((2, *IMAGE_SHAPE))

    def test_smoke_xr(self):
        def _test(shape: tuple[int, ...]):
            _, _, _, bi4, bi5 = get_data_xr(shape)
            cloud.vifcm_night(bi4, bi5)

        _test(IMAGE_SHAPE)
        _test((2, *IMAGE_SHAPE))
        _test((2, *IMAGE_SHAPE))

    @pytest.mark.parametrize(
        ("bi4", "bi5", "expected"),
        [
            (
                # bi4
                [np.nan, 100, 300, 200, 300],
                # bi5
                [np.nan, 100, 200, 300, 300],
                # expected
                [np.nan, 0, 1, 1, 1],
            )
        ],
    )
    def test_alg(self, bi4, bi5, expected):
        mask = cloud.vifcm_night(get_np_from_list(bi4), get_np_from_list(bi5))
        assert np.allclose(mask, get_np_from_list(expected), equal_nan=True)

        mask = cloud.vifcm_night(get_np_seq_from_list(bi4), get_np_seq_from_list(bi5))
        assert np.allclose(mask, get_np_seq_from_list(expected), equal_nan=True)

        mask = cloud.vifcm_night(get_xr_from_list(bi4), get_xr_from_list(bi5))
        assert mask.equals(get_xr_from_list(expected))

        mask = cloud.vifcm_night(get_xr_seq_from_list(bi4), get_xr_seq_from_list(bi5))
        assert mask.equals(get_xr_seq_from_list(expected))



import numpy as np
import pytest

from tests.algs.utils import (
    IMAGE_SHAPE,
    get_data_np,
    get_data_xr,
    get_np_from_list,
    get_np_seq_from_list,
    get_xr_from_list,
    get_xr_seq_from_list,
)
from viirs_tools.algs import night, utils


class TestMerge:
    def test_smoke_np(self):
        def _test(shape: tuple[int, ...]):
            ri1, _, _, bi4, _ = get_data_np(shape)
            nmask = night.naive(ri1, bi4)
            utils.merge_day_night(ri1, bi4, nmask)

        _test(IMAGE_SHAPE)
        _test((2, *IMAGE_SHAPE))

    def test_smoke_xr(self):
        def _test(shape: tuple[int, ...]):
            ri1, _, _, bi4, _ = get_data_xr(shape)
            nmask = night.naive(ri1, bi4)
            utils.merge_day_night(ri1, bi4, nmask)

        _test(IMAGE_SHAPE)
        _test((2, *IMAGE_SHAPE))

    @pytest.mark.parametrize(
        ("day", "night", "nmask", "expected"),
        [
            (
                # day
                [1, 1, 1],
                # night
                [5, 5, 5],
                # nmask
                [np.nan, 0, 1],
                # expected
                [5, 1, 5],
            )
        ],
    )
    def test_alg(self, day, night, nmask, expected):
        mask = utils.merge_day_night(get_np_from_list(day), get_np_from_list(night), get_np_from_list(nmask))
        assert np.allclose(mask, get_np_from_list(expected), equal_nan=True)

        mask = utils.merge_day_night(get_np_seq_from_list(day), get_np_seq_from_list(night), get_np_seq_from_list(nmask))
        assert np.allclose(mask, get_np_seq_from_list(expected), equal_nan=True)

        mask = utils.merge_day_night(get_xr_from_list(day), get_xr_from_list(night), get_xr_from_list(nmask))
        assert mask.equals(get_xr_from_list(expected))

        mask = utils.merge_day_night(get_xr_seq_from_list(day), get_xr_seq_from_list(night), get_xr_seq_from_list(nmask))
        assert mask.equals(get_xr_seq_from_list(expected))



import xarray as xr

from viirs_tools.utils.types import ArrayLike, _check_data


def merge_day_night(day: ArrayLike, night: ArrayLike, nmask: ArrayLike) -> ArrayLike:
    """Merge day and night composits

    Args:
        day_cm : day composit
        night_cm : night composit
        nmask : binary night mask

    Returns:
        Merged composit
    """
    _check_data(day, night, nmask)

    return xr.where(nmask == 0, day, night)



import numpy as np
import pytest

from tests.algs.utils import (
    IMAGE_SHAPE,
    get_data_np,
    get_data_xr,
    get_np_from_list,
    get_np_seq_from_list,
    get_xr_from_list,
    get_xr_seq_from_list,
)
from viirs_tools.algs import index


class TestNdvi:
    def test_smoke_np(self):
        def _test(shape: tuple[int, ...]):
            ri1, ri2, _, _, _ = get_data_np(shape)
            index.ndvi(ri2, ri1)

        _test(IMAGE_SHAPE)
        _test((2, *IMAGE_SHAPE))

    def test_smoke_xr(self):
        def _test(shape: tuple[int, ...]):
            ri1, ri2, _, _, _ = get_data_xr(shape)
            index.ndvi(ri2, ri1)

        _test(IMAGE_SHAPE)
        _test((2, *IMAGE_SHAPE))

    @pytest.mark.parametrize(
        ("ri1", "ri2", "expected"),
        [
            (
                # ri1
                [np.nan, 5, 5],
                # ri2
                [np.nan, 5, 15],
                # expected
                [np.nan, 0, 0.5],
            )
        ],
    )
    def test_alg(self, ri1, ri2, expected):
        mask = index.ndvi(get_np_from_list(ri2), get_np_from_list(ri1))
        assert np.allclose(mask, get_np_from_list(expected), equal_nan=True)

        mask = index.ndvi(get_np_seq_from_list(ri2), get_np_seq_from_list(ri1))
        assert np.allclose(mask, get_np_seq_from_list(expected), equal_nan=True)

        mask = index.ndvi(get_xr_from_list(ri2), get_xr_from_list(ri1))
        assert mask.equals(get_xr_from_list(expected))

        mask = index.ndvi(get_xr_seq_from_list(ri2), get_xr_seq_from_list(ri1))
        assert mask.equals(get_xr_seq_from_list(expected))


class TestNdsi:
    def test_smoke_np(self):
        def _test(shape: tuple[int, ...]):
            ri1, _, ri3, _, _ = get_data_np(shape)
            index.ndsi(ri1, ri3)

        _test(IMAGE_SHAPE)
        _test((2, *IMAGE_SHAPE))

    def test_smoke_xr(self):
        def _test(shape: tuple[int, ...]):
            ri1, _, ri3, _, _ = get_data_xr(shape)
            index.ndsi(ri1, ri3)

        _test(IMAGE_SHAPE)
        _test((2, *IMAGE_SHAPE))

    @pytest.mark.parametrize(
        ("ri1", "ri3", "expected"),
        [
            (
                # ri1
                [np.nan, 5, 15],
                # ri3
                [np.nan, 5, 5],
                # expected
                [np.nan, 0, 0.5],
            )
        ],
    )
    def test_alg(self, ri1, ri3, expected):
        mask = index.ndvi(get_np_from_list(ri1), get_np_from_list(ri3))
        assert np.allclose(mask, get_np_from_list(expected), equal_nan=True)

        mask = index.ndvi(get_np_seq_from_list(ri1), get_np_seq_from_list(ri3))
        assert np.allclose(mask, get_np_seq_from_list(expected), equal_nan=True)

        mask = index.ndvi(get_xr_from_list(ri1), get_xr_from_list(ri3))
        assert mask.equals(get_xr_from_list(expected))

        mask = index.ndvi(get_xr_seq_from_list(ri1), get_xr_seq_from_list(ri3))
        assert mask.equals(get_xr_seq_from_list(expected))



import numpy as np
from netCDF4 import Dataset  # require netcdf4 being installed, not NetCDF4
from numpy import ma


def _get_masked(data: np.ndarray, thr: float) -> ma.MaskedArray:
    """Do masking on the given array for specified threshold

    Args:
        data : given data array
        thr : threshold value

    Returns:
        resulting array
    """

    mask = data >= thr if thr > 0 else data <= thr
    return ma.masked_array(data, mask=mask)


def read_so_data(name: str, file: Dataset) -> ma.MaskedArray:
    """Read from NASA distributed hdf's and nc's files
    data stored in Scale-Offset model

    Args:
        name : name of the desired dataset
        file : file-like object, created with netCDF4

    Returns:
        dataset from the file
    """
    scale = 1
    offset = 0
    if "Scale" in file.variables[name].ncattrs():
        scale = file.variables[name].getncattr("Scale")
    if "Offset" in file.variables[name].ncattrs():
        offset = file.variables[name].getncattr("Offset")
    ref = file.variables[name][:]
    thr = file.variables[name].getncattr("FILL_TEST_VALUE").split("=")[1]
    return _get_masked(ref, float(thr)) * scale + offset


# Wrappers for handy extracting different types of data


def read_ref(band: str, file: Dataset) -> ma.MaskedArray:
    return read_so_data(f"Reflectance_{band}", file)


def read_rad(band: str, file: Dataset) -> ma.MaskedArray:
    return read_so_data(f"Radiance_{band}", file)


def read_bt(band: str, file: Dataset) -> ma.MaskedArray:
    return read_so_data(f"BrightnessTemperature_{band}", file)


def read_lat(file: Dataset) -> ma.MaskedArray:
    return read_so_data("Latitude", file)


def read_lon(file: Dataset) -> ma.MaskedArray:
    return read_so_data("Longitude", file)


def read_solza(file: Dataset) -> ma.MaskedArray:
    return read_so_data("SolarZenithAngle", file)


def read_solaa(file: Dataset):
    return read_so_data("SolarAzimuthAngle", file)


def read_satza(file: Dataset):
    return read_so_data("SatelliteZenithAngle", file)


def read_sataa(file: Dataset):
    return read_so_data("SatelliteAzimuthAngle", file)



import numpy as np
import xarray as xr

from viirs_tools.utils.types import ArrayLike, _check_data


def naive(refband: ArrayLike, btband: ArrayLike) -> ArrayLike:
    """Get night mask from any reflectance and brightness bands
    Day/Night state here meets the condition SZA < 90 deg
    Assumed that data was loaded in the reflectance or
        brightness-temperature, not radiance calibration

    Args:
        refband : any reflectance data
        btband : any brightness-temperature data

    Returns:
        Integer mask, 1 means night state, 0 means day state,
            Can contain NaN values in case of missing data in BT band
    """
    _check_data(refband, btband)

    bmask = ~xr.ufuncs.isnan(btband)
    rmask = xr.ufuncs.isnan(refband)
    return xr.where(bmask, rmask, np.nan)



from enum import Enum
from typing import Any

import numpy as np
import xarray as xr

ArrayLike = np.ndarray | xr.DataArray


def _check_data(*args: Any):
    if len(args) == 0:
        return True

    if type(args[0]) is not ArrayLike:
        return False

    if len(args) == 1:
        return True

    target_type = type(args[0])
    target_shape = args[0].shape

    return all(not (type(arg) is not target_type or arg.shape != target_shape) for arg in args[1:])


class AlgEnum(Enum):
    pass



from viirs_tools.runner import AlgsCloud, AlgsIndex, AlgsLST, AlgsNight, AlgsUtils, AlgsWater, Runner

__all__ = ["Runner", "AlgsCloud", "AlgsIndex", "AlgsLST", "AlgsNight", "AlgsUtils", "AlgsWater"]



import pytest

from viirs_tools.runner import AlgsCloud, AlgsIndex, AlgsLST, AlgsNight, AlgsUtils, AlgsWater, Runner


class TestRunner:
    @pytest.mark.parametrize(
        "algs",
        [
            AlgsCloud,
            AlgsIndex,
            AlgsLST,
            AlgsNight,
            AlgsWater,
            AlgsUtils,
        ],
    )
    def test_integrity(self, algs):
        runner = Runner()
        impls = runner._IMPLS[algs]
        for alg in algs:
            assert alg in impls
        assert len(algs) == len(impls)



import numpy as np
import pytest

from tests.algs.utils import (
    IMAGE_SHAPE,
    get_data_np,
    get_data_xr,
    get_np_from_list,
    get_np_seq_from_list,
    get_xr_from_list,
    get_xr_seq_from_list,
)
from viirs_tools.algs import night


class TestNaive:
    def test_smoke_np(self):
        def _test(shape: tuple[int, ...]):
            ri1, _, _, bi4, _ = get_data_np(shape)
            night.naive(ri1, bi4)

        _test(IMAGE_SHAPE)
        _test((2, *IMAGE_SHAPE))

    def test_smoke_xr(self):
        def _test(shape: tuple[int, ...]):
            ri1, _, _, bi4, _ = get_data_xr(shape)
            night.naive(ri1, bi4)

        _test(IMAGE_SHAPE)
        _test((2, *IMAGE_SHAPE))

    @pytest.mark.parametrize(
        ("ref", "bt", "expected"),
        [
            (
                # ref
                [np.nan, np.nan, 20, 30],
                # bt
                [np.nan, 40, np.nan, 1],
                # expected
                [np.nan, 1, np.nan, 0],
            )
        ],
    )
    def test_alg(self, ref, bt, expected):
        mask = night.naive(get_np_from_list(ref), get_np_from_list(bt))
        assert np.allclose(mask, get_np_from_list(expected), equal_nan=True)

        mask = night.naive(get_np_seq_from_list(ref), get_np_seq_from_list(bt))
        assert np.allclose(mask, get_np_seq_from_list(expected), equal_nan=True)

        mask = night.naive(get_xr_from_list(ref), get_xr_from_list(bt))
        assert mask.equals(get_xr_from_list(expected))

        mask = night.naive(get_xr_seq_from_list(ref), get_xr_seq_from_list(bt))
        assert mask.equals(get_xr_seq_from_list(expected))



from tests.algs.utils import IMAGE_SHAPE, get_data_np, get_data_xr
from viirs_tools.algs import index, lst


class TestLst:
    def test_smoke_np(self):
        def _test(shape: tuple[int, ...]):
            ri1, ri2, _, _, bi05 = get_data_np(shape)
            ndvi = index.ndvi(ri2, ri1)
            lst.mono_window_i05(bi05, ndvi)

        _test(IMAGE_SHAPE)
        _test((2, *IMAGE_SHAPE))

    def test_smoke_xr(self):
        def _test(shape: tuple[int, ...]):
            ri1, ri2, _, _, bi05 = get_data_xr(shape)
            ndvi = index.ndvi(ri2, ri1)
            lst.mono_window_i05(bi05, ndvi)

        _test(IMAGE_SHAPE)
        _test((2, *IMAGE_SHAPE))



import numpy as np
import xarray as xr

from viirs_tools.utils.types import ArrayLike, _check_data


def water_bodies_day(ri1: ArrayLike, ri2: ArrayLike, ri3: ArrayLike) -> ArrayLike:
    """Day reflectance water bodies test
    Based on the W.Schroeder, P.Oliva, L.Giglio, I.A.Csiszar (2014).
    The New VIIRS 375 m active fire detection data product:
        Algorithm description and initial assessment

    Args:
        ri1 : I01 in reflectance calibration
        ri2 : I02 in reflectance calibration
        ri3 : I03 in reflectance calibration

    Returns:
        Binary water bodies mask, 0 is clear water body, 1 is clear pixel
            Can contain NaN values
    """
    _check_data(ri1, ri2, ri3)

    mask = (ri1 > ri2) & (ri2 > ri3)

    return 1 - xr.where(xr.ufuncs.isnan(ri1), np.nan, mask)



import os
import subprocess as sp
from collections.abc import Callable
from concurrent.futures import ProcessPoolExecutor
from datetime import datetime, timedelta


def _test_cmrfetch():
    """Test for availability of cmrfetch
    Test was unsuccessfull in the case of any exception
    """
    cf = "cmrfetch"
    try:
        sp.run([cf, "--version"], check=True, stdout=sp.DEVNULL, stderr=sp.DEVNULL)
    except FileNotFoundError as e:
        raise Exception(f"{cf} is not accessible") from e
    except sp.CalledProcessError as e:
        if e.returncode != 0:
            raise Exception(f"{cf} was found, failed to execute") from e


def _get_data_for_interval(
    path: str,
    start_d: str,
    end_d: str,
    names: list[str],
    geobox: str,
    dconc: int = 4,
):
    """Worker function for downloading data itself

    Args:
        path : path for saving downloaded files
        start_d : start of the time selection interval (including)
        end_d : end of the time selection interval (excluding)
        names : shortnames for desired collections
        geobox : bounding box for selecting granules for format information check cmrfetch docs
        dconc : number of concurrently downloading connections
    """
    snames = ""
    for i in names:
        snames += f"-s {i} "
    x = 1
    cnt = 0
    while x != 0:
        # really rough solution, timeouts are handled by cmrfetch
        # so they cannot be recognized as runtime error below
        # TODO: cmrfetch output analysis for re-downloading failed files
        fetch_cli = [
            "cmrfetch",
            "granules",
            str(snames),
            "-t",
            str(start_d),
            ",",
            str(end_d),
            "--bounding-box",
            str(geobox),
            "--download",
            str(path),
            "--download-concurrency",
            str(dconc),
        ]
        r = sp.run(fetch_cli, check=True)
        x = r.returncode
        cnt += 1
        if cnt == 5:
            raise RuntimeError(f"Cannot download data for {start_d},{end_d}")


def assimilate(
    names: list[str],
    geobox: str,
    start_d: datetime,
    end_d: datetime,
    path: str,
    workers: int,
    max_queue: int = 0,
    assim_callback: Callable[[str], None] = (lambda path: None),
    dconc: int = 4,
):
    """Performs data assimilation - download it using cmrfetch
    per day intervals and then perform callback function
    on it (for compression purposes mainly)

    Args:
        names : shortnames for desired collections
        geobox: bounding box for selecting granules
            for format information check cmrfetch docs
        start_d : start of the time selection interval (including)
        end_d : end of the time selection interval (excluding)
        path : path for saving downloaded files
        workers : max num of the processes performing callbacks
        max_queue : max num of the queued callback
            tasks, used for downloading throttling
            (for avoiding disk filling)
        assim_callback : callback to perform
            on the each set of data per single day
        dconc : number of concurrently downloading connections
    """
    _test_cmrfetch()

    if max_queue == 0:
        max_queue = 2 * workers

    if not os.path.exists(path):
        os.mkdir(path)

    with ProcessPoolExecutor(max_workers=workers) as executor:
        futures = []
        cur_d = end_d
        next_d = end_d + timedelta(days=1)
        while cur_d >= start_d:
            cur_d_s = cur_d.strftime("%Y-%m-%d")
            next_d_s = next_d.strftime("%Y-%m-%d")
            step_path = os.path.join(path, f"{cur_d_s}")

            _get_data_for_interval(step_path, cur_d_s, next_d_s, names, geobox, dconc)

            task = executor.submit(assim_callback, step_path)
            futures.append(task)

            next_d = cur_d
            cur_d -= timedelta(days=1)

            # Throttling data downloading
            if len(futures) > max_queue:
                while len(futures) > max_queue:
                    futures = [f for f in futures if not f.done()]
                    if len(futures) != 0:
                        futures[0].result()
